ENV_DATABASE_URL : str = 'DATABASE_URL'
ENV_MQTT_HOST    : str = 'MQTT_HOST'
ENV_MQTT_PORT    : str = 'MQTT_PORT'
ENV_UPLOAD_PATH  : str = 'UPLOAD_PATH'

TOPIC_ACTION       : str = 'espark/action'
TOPIC_DEVICE       : str = 'espark/device'
TOPIC_REGISTRATION : str = 'espark/registration'
TOPIC_TELEMETRY    : str = 'espark/telemetry'
TOPIC_OTA          : str = 'espark/ota'
TOPIC_CRASH        : str = 'espark/crash'



from ..models import AppVersion
from .base_repository import AsyncRepository


class AppVersionRepository(AsyncRepository[AppVersion]):
    def __init__(self):
        super().__init__(AppVersion)

    async def get_by_app_name(self, session, app_name: str) -> AppVersion | None:
        # pylint: disable=unexpected-keyword-arg
        return await self.get(session, AppVersion.id == app_name)



from datetime import datetime


def log_debug(message: str) -> None:
    print(f'{datetime.now().isoformat()} [DEBUG] {message}')


def log_error(e: Exception) -> None:
    print(f'{datetime.now().isoformat()} [ERROR] {type(e).__name__}: {e}')

    raise e



from abc import ABC, abstractmethod


class BaseNotifier(ABC):
    @abstractmethod
    async def notify(self, device_id: str, event_type: str, value: int) -> None:
        """
        Send a notification for an event.

        Args:
            device_id: ID of the device generating the event.
            event_type: Type of the event (human presence, door open, etc).
            value: An integer value associated with the event.
        """



from ..data.models import Trigger
from ..data.repositories import TriggerRepository
from .base_router import BaseRouter


class TriggerRouter(BaseRouter):
    def __init__(self, repo: TriggerRepository = None) -> None:
        self.repo : TriggerRepository = repo or TriggerRepository()

        super().__init__(Trigger, self.repo, '/api/v1/triggers', ['trigger'])



from typing import Dict

from sqlalchemy import Column
from sqlmodel import SQLModel, Field, JSON


class Notification(SQLModel, table=True):
    id       : int            = Field(primary_key=True, description='Unique identifier for the notification')
    name     : str            = Field(index=True, unique=True, description='Name of the notification')
    provider : str            = Field(index=True, description='Notification provider (e.g., Slack, Twilio)')
    config   : Dict[str, str] = Field(default_factory=dict, sa_column=Column(JSON), description='JSON string of provider-specific configuration parameters')



from typing import Optional

from sqlmodel import Field, SQLModel


class Trigger(SQLModel, table=True):
    id               : Optional[int] = Field(primary_key=True, description='Unique identifier for the trigger')
    name             : str           = Field(index=True, unique=True, description='Name of the trigger')
    device_id        : Optional[str] = Field(index=True, default=None, description='Identifier of the associated device')
    data_type        : Optional[str] = Field(index=True, default=None, description='Type of telemetry data the trigger monitors (e.g., temperature, humidity)')
    condition        : str           = Field(description='Condition to evaluate (e.g., ">", "<=")')
    value            : int           = Field(description='Value to compare against for the trigger condition')
    notification_ids : Optional[str] = Field(default=None, description='Comma-separated list of associated notification IDs to be sent when the trigger condition is met')



from datetime import datetime
from typing import Dict, Optional

from sqlalchemy import Column
from sqlmodel import SQLModel, Field, JSON


class Device(SQLModel, table=True):
    id           : str                         = Field(primary_key=True, description='Unique identifier for the device')
    display_name : Optional[str]               = Field(default=None, description='Human-readable name of the device')
    app_name     : Optional[str]               = Field(default=None, foreign_key='appversion.id', ondelete='CASCADE', description='Name of the application running on the device')
    app_version  : Optional[str]               = Field(default=None, description='Version of the application running on the device')
    capabilities : Optional[str]               = Field(default=None, description='Comma separated capabilities of the device')
    parameters   : Dict[str, str | int | bool] = Field(default_factory=dict, sa_column=Column(JSON), description='JSON string of capability-specific parameters')
    last_seen    : datetime                    = Field(index=True, description='Last time the device was seen online')



from ..data.models import AppVersion
from ..data.repositories import AppVersionRepository
from .base_router import BaseRouter


class AppVersionRouter(BaseRouter):
    def __init__(self, repo: AppVersionRepository = None) -> None:
        self.repo : AppVersionRepository = repo or AppVersionRepository()

        super().__init__(AppVersion, self.repo, '/api/v1/apps', ['version'])



from datetime import datetime
from typing import Optional

from sqlmodel import SQLModel, Field


class Telemetry(SQLModel, table=True):
    id        : Optional[int] = Field(primary_key=True, default=None)
    device_id : str           = Field(foreign_key='device.id', ondelete='CASCADE', description='Device that sent this data')
    timestamp : datetime      = Field(index=True, description='Timestamp of the data')
    data_type : str           = Field(index=True, description='Type of the data (e.g., motion, temperature)')
    value     : int           = Field(description='Value of the data (e.g., temperature, human presence detected or not)')



from json import dumps
from os import getenv

from aiomqtt import Client

from ..constants import ENV_MQTT_HOST, ENV_MQTT_PORT, TOPIC_ACTION
from ..data import async_session
from ..data.repositories import OutboxRepository


async def consume_outbox(device_id: str, event_type: str) -> None:
    outbox_repo = OutboxRepository()

    async with async_session() as session:
        event = await outbox_repo.get_next(session, device_id, event_type)
        if not event:
            return

        await outbox_repo.delete_pending(session, device_id, event_type)

        async with Client(getenv(ENV_MQTT_HOST, 'localhost'), int(getenv(ENV_MQTT_PORT, '1883'))) as client:
            await client.publish(f'{TOPIC_ACTION}/{device_id}', dumps(event.payload), qos=1, retain=True)

        await session.commit()



from datetime import datetime
from typing import Dict, Optional
from uuid import UUID, uuid4

from sqlalchemy import Column
from sqlmodel import SQLModel, Field, JSON, UniqueConstraint


class OutboxEvent(SQLModel, table=True):
    __table_args__ = (
        UniqueConstraint('device_id', 'event_type', 'is_processed', name='uq_device_event'),
    )

    id           : UUID               = Field(default_factory=uuid4, primary_key=True, description='Unique identifier for the outbox event')
    device_id    : str                = Field(foreign_key='device.id', ondelete='CASCADE', index=True, description='Identifier of the device associated with the event')
    event_type   : str                = Field(index=True, description='Type of the event')
    payload      : Dict[str, str]     = Field(default_factory=dict, sa_column=Column(JSON), description='JSON string of event-specific payload data')
    created_at   : datetime           = Field(default_factory=datetime.now, index=True, description='Timestamp when the event was created')
    processed_at : Optional[datetime] = Field(default=None, index=True, description='Timestamp when the event was processed')
    is_processed : bool               = Field(default=False, index=True, description='Flag indicating whether the event has been processed')



from .device_router import DeviceRouter
from .notification_router import NotificationRouter
from .telemetry_router import TelemetryRouter
from .trigger_router import TriggerRouter
from .version_router import AppVersionRouter



from json import dumps
from os import getenv

from aiomqtt import Client
from fastapi import Depends, Query, Response
from sqlalchemy.ext.asyncio import AsyncSession

from ..constants import ENV_MQTT_HOST, ENV_MQTT_PORT, TOPIC_DEVICE
from ..data.models import Device
from ..data.repositories import DeviceRepository, TelemetryRepository
from ..utils import log_debug
from .base_router import BaseRouter


class DeviceRouter(BaseRouter):
    def __init__(self, repo: DeviceRepository = None) -> None:
        self.repo : DeviceRepository = repo or DeviceRepository()

        super().__init__(Device, self.repo, '/api/v1/devices', ['device'])

    async def _publish_update(self, entity: Device) -> None:
        log_debug(f'Publishing parameters update for device {entity.id}: {entity.parameters}')

        async with Client(getenv(ENV_MQTT_HOST, 'localhost'), int(getenv(ENV_MQTT_PORT, '1883'))) as client:
            await client.publish(f'{TOPIC_DEVICE}/{entity.id}', payload=dumps(entity.parameters) if entity.parameters else None, qos=1, retain=True)

    async def _after_update(self, entity: Device, session: AsyncSession) -> None:
        await super()._after_update(entity, session)

        await self._publish_update(entity)

    def _setup_routes(self) -> None:
        @self.router.get('/all')
        async def list_all(response: Response, session: AsyncSession = Depends(BaseRouter._get_session), order_by: str = Query(None), offset: int = Query(0, ge=0), limit: int = Query(10, ge=1, le=100)):
            response.headers['X-Total-Count'] = str(await self.repo.count(session))

            devices                = await self.repo.list(session, order_by=order_by, offset=offset, limit=limit)
            devices_with_telemetry = []
            telemetry_repo         = TelemetryRepository()

            for device in devices:
                battery = await telemetry_repo.get_latest_for_device(session, device.id, 'battery')

                devices_with_telemetry.append({
                    'id'           : device.id,
                    'display_name' : device.display_name,
                    'app_name'     : device.app_name,
                    'app_version'  : device.app_version,
                    'capabilities' : device.capabilities,
                    'parameters'   : device.parameters,
                    'last_seen'    : device.last_seen,
                    'battery'      : battery.value if battery else None,
                })

            return devices_with_telemetry

        super()._setup_routes()



from typing import Generic, Optional, Sequence, Type, TypeVar

from sqlalchemy import ColumnElement, func, text
from sqlalchemy.ext.asyncio import AsyncSession
from sqlmodel import select, SQLModel

T = TypeVar('T', bound=SQLModel)


class AsyncRepository(Generic[T]):
    def __init__(self, model: Type[T]):
        self.model = model

    async def count(self, session: AsyncSession, *conditions: ColumnElement) -> int:
        # pylint: disable=not-callable
        query = select(func.count()).select_from(self.model)

        if conditions:
            query = query.where(*conditions)

        result = await session.execute(query)
        return result.scalar_one()

    async def add(self, session: AsyncSession, entity: T) -> T:
        session.add(entity)

        await session.commit()
        await session.refresh(entity)

        return entity

    async def delete(self, session: AsyncSession, entity: T) -> None:
        await session.delete(entity)
        await session.commit()

    async def get(self, session: AsyncSession, *conditions: ColumnElement) -> Optional[T]:
        query = select(self.model)

        if conditions:
            query = query.where(*conditions)

        return (await session.execute(query)).scalars().first()

    async def list(self, session: AsyncSession, *conditions: ColumnElement[bool], offset: Optional[int] = None, order_by: ColumnElement | str = None, limit: Optional[int] = None) -> Sequence[T]:
        query = select(self.model)

        if conditions:
            query = query.where(*conditions)

        if order_by is not None:
            query = query.order_by(text(order_by) if isinstance(order_by, str) else order_by)

        if offset is not None:
            query = query.offset(offset)

        if limit is not None:
            query = query.limit(limit)

        return (await session.execute(query)).scalars().all()

    async def update(self, session: AsyncSession, entity: T, **kwargs) -> T:
        for key, value in kwargs.items():
            setattr(entity, key, value)

        session.add(entity)

        await session.commit()
        await session.refresh(entity)

        return entity



from slack_sdk import WebClient

from ..utils import log_debug
from .notifier import BaseNotifier


class SlackNotifier(BaseNotifier):
    def __init__(self, slack_token: str, slack_channel: str) -> None:
        self.slack_token   = slack_token
        self.slack_channel = slack_channel

    async def notify(self, device_id: str, event_type: str, value: int) -> None:
        client = WebClient(token=self.slack_token)

        log_debug(f'Posting {event_type} event to Slack for device {device_id}: {value}')

        client.chat_postMessage(channel=self.slack_channel, text=f'Device {device_id} reported {event_type} with value {value}.')



from .base_repository import AsyncRepository
from .device_repository import DeviceRepository
from .notification_repository import NotificationRepository
from .outbox_repository import OutboxRepository
from .telemetry_repository import TelemetryRepository
from .trigger_repository import TriggerRepository
from .version_repository import AppVersionRepository



from ..models import Notification
from .base_repository import AsyncRepository


class NotificationRepository(AsyncRepository[Notification]):
    def __init__(self):
        super().__init__(Notification)



from asyncio import create_task, gather, Queue, sleep
from datetime import datetime, timezone
from json import dumps, JSONDecodeError, loads
from os import getenv
from uuid import getnode

from aiomqtt import Client, MqttError
from packaging.version import Version

from ..constants import ENV_MQTT_HOST, ENV_MQTT_PORT, TOPIC_CRASH, TOPIC_OTA, TOPIC_REGISTRATION, TOPIC_TELEMETRY
from ..data import async_session
from ..data.models import AppVersion, Device, Telemetry
from ..data.repositories import AppVersionRepository, DeviceRepository, NotificationRepository, TelemetryRepository, TriggerRepository
from ..notifications import SlackNotifier
from ..utils import log_debug, log_error

MQTT_RETRY_DELAY: int = 5


class MQTTManager:
    def __init__(self, version_repo: AppVersionRepository = None, device_repo: DeviceRepository = None, notification_repo: NotificationRepository = None, telemetry_repo: TelemetryRepository = None, trigger_repo: TriggerRepository = None) -> None:
        self.mqtt_host         : str                  = getenv(ENV_MQTT_HOST, 'localhost')
        self.mqtt_port         : int                  = int(getenv(ENV_MQTT_PORT, '1883'))
        self.version_repo      : AppVersionRepository = version_repo if version_repo else AppVersionRepository()
        self.device_repo       : DeviceRepository     = device_repo if device_repo else DeviceRepository()
        self.notification_repo : NotificationRepository = notification_repo if notification_repo else NotificationRepository()
        self.telemetry_repo    : TelemetryRepository  = telemetry_repo if telemetry_repo else TelemetryRepository()
        self.trigger_repo      : TriggerRepository    = trigger_repo if trigger_repo else TriggerRepository()
        self.queue             : Queue                = Queue()

    async def _handle_registration(self, device_id: str, payload: dict) -> None:
        try:
            log_debug(f'Registering device: {device_id}')

            async with async_session() as session:
                device = await self.device_repo.get(session, Device.id == device_id)
                if device:
                    await self.device_repo.update(session, device, last_seen=datetime.now(timezone.utc))
                else:
                    device = Device()

                    device.id           = device_id
                    device.display_name = None
                    device.app_name     = payload['app_name']
                    device.app_version  = payload['app_version']
                    device.capabilities = payload['capabilities']
                    device.last_seen    = datetime.now(timezone.utc)

                    await self.device_repo.add(session, device)

                latest_version = await self.version_repo.get(session, AppVersion.id == payload['app_name'])
                if not latest_version:
                    latest_version = AppVersion()
                    latest_version.id      = payload['app_name']
                    latest_version.version = payload['app_version']

                    await self.version_repo.add(session, latest_version)

                current_version = Version(payload['app_version'])

                if Version(latest_version.version) > current_version:
                    log_debug(f'Device {device_id} is running an outdated version ({current_version} < {latest_version.version})')

                    async with Client(self.mqtt_host, self.mqtt_port, identifier=f'espark-core-{hex(getnode())}') as client:
                        await client.publish(f'{TOPIC_OTA}/{device_id}', dumps({
                            'device_id'    : device_id,
                            'app_name'     : payload['app_name'],
                            'app_version'  : latest_version.version,
                            'download_url' : f'/downloads/{payload["app_name"]}/{latest_version.version}',
                        }), qos=1)
        # pylint: disable=broad-exception-caught
        except Exception as e:
            log_error(e)

    async def _handle_telemetry(self, device_id: str, payload: dict) -> None:
        try:
            log_debug(f'Receiving telemetry from device: {device_id} - Payload: {payload}')

            async with async_session() as session:
                telemetry = Telemetry()

                telemetry.device_id = device_id
                telemetry.timestamp = datetime.now(timezone.utc)
                telemetry.data_type = payload.get('data_type')
                telemetry.value     = payload.get('value')

                await self.telemetry_repo.add(session, telemetry)
        # pylint: disable=broad-exception-caught
        except Exception as e:
            log_error(e)

        await self._handle_triggers(device_id, payload.get('data_type'), payload.get('value'))

    async def _handle_triggers(self, device_id: str, data_type: str, value: int) -> None:
        async with async_session() as session:
            triggers = await self.trigger_repo.list(session)
            for trigger in triggers:
                matched_device_id = trigger.device_id is None or trigger.device_id == device_id
                matched_data_type = trigger.data_type is None or trigger.data_type == data_type
                if matched_device_id and matched_data_type:
                    condition_met = False
                    if trigger.condition is None:
                        condition_met = True
                    elif trigger.condition == '==' and value == trigger.value:
                        condition_met = True
                    elif trigger.condition == '>' and value > trigger.value:
                        condition_met = True
                    elif trigger.condition == '>=' and value >= trigger.value:
                        condition_met = True
                    elif trigger.condition == '<' and value < trigger.value:
                        condition_met = True
                    elif trigger.condition == '<=' and value <= trigger.value:
                        condition_met = True

                    if condition_met:
                        log_debug(f'Trigger {trigger.name} activated for device {device_id} with data type {data_type} and value {value}')

                        notifications = await self.notification_repo.list(session)
                        for notification in notifications:
                            if str(notification.id) in trigger.notification_ids:
                                log_debug(f'Sending notification {notification.name} for trigger {trigger.name}')

                                if notification.provider == 'Slack':
                                    await SlackNotifier(notification.config['slack_token'], notification.config['slack_channel']).notify(device_id, data_type, value)

    async def _process_queue(self) -> None:
        while True:
            try:
                topic, payload = await self.queue.get()

                topic_parts: list[str] = str(topic).split('/')
                if len(topic_parts) != 3:
                    log_error(Exception('Invalid topic format, skipping message'))
                    continue

                app_type, message_type, device_id = topic_parts
                if app_type != 'espark':
                    log_debug('Invalid app type, skipping message')
                    continue

                try:
                    payload = loads(payload.decode())
                except JSONDecodeError as e:
                    log_error(e)
                    continue

                if message_type == TOPIC_REGISTRATION.split('/')[1]:
                    await self._handle_registration(device_id, payload)
                elif message_type == TOPIC_TELEMETRY.split('/')[1]:
                    await self._handle_telemetry(device_id, payload)
                elif message_type == TOPIC_CRASH.split('/')[1]:
                    log_debug(f'Received crash report from device {device_id}: {payload}')
                else:
                    log_debug(f'Unknown message type "{message_type}", skipping message')

                self.queue.task_done()
            # pylint: disable=broad-exception-caught
            except Exception as e:
                log_error(e)

    async def _process_messages(self) -> None:
        while True:
            try:
                async with Client(self.mqtt_host, self.mqtt_port, identifier=f'espark-core-{hex(getnode())}') as client:
                    await client.subscribe(f'{TOPIC_REGISTRATION}/+')
                    await client.subscribe(f'{TOPIC_TELEMETRY}/+')
                    await client.subscribe(f'{TOPIC_CRASH}/+')

                    async for message in client.messages:
                        log_debug(f'Received MQTT message on topic {message.topic}')

                        self.queue.put_nowait((message.topic, message.payload))
            except MqttError as e:
                log_error(e)

                await sleep(MQTT_RETRY_DELAY)

    async def start(self) -> None:
        await gather(create_task(self._process_queue()), create_task(self._process_messages()))



from ..data.models import Notification
from ..data.repositories import NotificationRepository
from .base_router import BaseRouter


class NotificationRouter(BaseRouter):
    def __init__(self, repo: NotificationRepository = None) -> None:
        self.repo : NotificationRepository = repo or NotificationRepository()

        super().__init__(Notification, self.repo, '/api/v1/notifications', ['notification'])



from os import getenv

from sqlalchemy.ext.asyncio import async_sessionmaker, AsyncSession, create_async_engine
from sqlmodel import SQLModel

from ..constants import ENV_DATABASE_URL

# pylint: disable=invalid-name
engine        = create_async_engine(getenv(ENV_DATABASE_URL, 'sqlite+aiosqlite:///database.db'), echo=True)
async_session = async_sessionmaker(engine, class_=AsyncSession, expire_on_commit=False)


async def init_db():
    async with engine.begin() as conn:
        await conn.run_sync(SQLModel.metadata.create_all)



from datetime import datetime, timedelta, timezone
from typing import Sequence

from fastapi import Depends, Query, Response
from sqlalchemy.ext.asyncio import AsyncSession
from sqlmodel import and_

from ..data.models import Telemetry
from ..data.repositories import DeviceRepository, TelemetryRepository
from .base_router import BaseRouter


class TelemetryRouter(BaseRouter):
    def __init__(self, repo: TelemetryRepository = None) -> None:
        self.repo : TelemetryRepository = repo or TelemetryRepository()

        super().__init__(Telemetry, self.repo, '/api/v1/telemetry', ['telemetry'])

    def _setup_routes(self) -> None:
        @self.router.get('/history', response_model=Sequence[Telemetry])
        async def list_history(response: Response, session: AsyncSession = Depends(BaseRouter._get_session), device_id: str = Query(..., min_length=1), offset: int = Query(0, min=0)) -> Sequence[Telemetry]:
            from_date = datetime.now(timezone.utc) - timedelta(seconds=offset)
            results   = await self.repo.list(session, and_(Telemetry.device_id == device_id, Telemetry.timestamp >= from_date))

            response.headers['X-Total-Count'] = str(len(results))

            return results

        @self.router.get('/recent', response_model=Sequence[Telemetry])
        async def list_recent(response: Response, session: AsyncSession = Depends(BaseRouter._get_session), offset: int = Query(0, min=0)) -> Sequence[Telemetry]:
            from_date   = datetime.now(timezone.utc) - timedelta(seconds=offset)
            device_repo = DeviceRepository()
            # pylint: disable=no-member
            devices     = await device_repo.list(session)
            results     = []

            for device in devices:
                capabilities = device.capabilities.split(',') if device.capabilities else []
                for data_type in capabilities:
                    if not data_type.startswith('action_'):
                        telemetry = await self.repo.get_latest_for_device(session, device.id, data_type)
                        if telemetry:
                            if telemetry.timestamp.tzinfo is None:
                                telemetry.timestamp = telemetry.timestamp.replace(tzinfo=timezone.utc)

                            if telemetry.timestamp >= from_date:
                                results.append(telemetry)

            response.headers['X-Total-Count'] = str(len(results))

            results.sort(key=lambda result: result.timestamp, reverse=True)

            return results

        super()._setup_routes()



from typing import Optional, Sequence

from sqlalchemy import ColumnElement
from sqlalchemy.ext.asyncio import AsyncSession
from sqlmodel import and_, select

from ..models import Telemetry
from .base_repository import AsyncRepository


class TelemetryRepository(AsyncRepository[Telemetry]):
    def __init__(self):
        super().__init__(Telemetry)

    async def get_latest_for_device(self, session: AsyncSession, device_id: str, data_type: str) -> Optional[Telemetry]:
        # pylint: disable=no-member
        results = await self.list(session, and_(Telemetry.device_id == device_id, Telemetry.data_type == data_type), order_by=Telemetry.timestamp.desc(), limit=1)
        return results[0] if results else None

    async def list(self, session: AsyncSession, *conditions: ColumnElement[bool], offset: Optional[int] = None, order_by: ColumnElement | str = None, limit: Optional[int] = None) -> Sequence[Telemetry]:
        if order_by is None:
            # pylint: disable=no-member
            order_by = Telemetry.timestamp.desc()

        return await super().list(session, *conditions, offset=offset, order_by=order_by, limit=limit)

    async def search(self, session: AsyncSession, device_id: str = None, data_type: str = None, condition: str = None, value: int = None):
        query = select(self.model)

        if device_id is not None:
            query = query.where(Telemetry.device_id == device_id)

        if data_type is not None:
            query = query.where(Telemetry.data_type == data_type)

        if condition is not None and value is not None:
            if condition == '<':
                query = query.where(Telemetry.value < value)
            elif condition == '<=':
                query = query.where(Telemetry.value <= value)
            elif condition == '==':
                query = query.where(Telemetry.value == value)
            elif condition == '>=':
                query = query.where(Telemetry.value >= value)
            elif condition == '>':
                query = query.where(Telemetry.value > value)

        return (await session.execute(query)).scalars().all()



from pathlib import Path
from setuptools import find_packages, setup

def read_requirements(filename: str):
    return [line.strip() for line in Path(filename).read_text().splitlines() if line.strip() and not line.startswith("#")]

root                    = Path(__file__).resolve().parent
production_requirements = (root / 'requirements.txt').read_text().splitlines()
extra_requirements      = (root / 'requirements.dev.txt').read_text().splitlines()

setup(
    name='espark-core',
    version='0.5.12',
    description='The core module of the Espark ESP32-based IoT device management framework.',
    long_description=(root / 'README.md').read_text(),
    long_description_content_type='text/markdown',
    license='MIT',
    license_files=[
        'LICENSE',
    ],
    python_requires='>=3.10',
    packages=find_packages(where='.'),
    package_dir={
        '' : '.',
    },
    install_requires=production_requirements,
    extras_require={
        'dev' :  extra_requirements,
    },
)



from sqlmodel import SQLModel, Field


class AppVersion(SQLModel, table=True):
    id      : str = Field(primary_key=True, description='The app name, which is a unique identifier for the app version')
    version : str = Field(description='Version of the application')



from datetime import datetime, timedelta, timezone
from typing import Optional, Sequence

from sqlalchemy.ext.asyncio import AsyncSession

from ..models import Device
from .base_repository import AsyncRepository

OFFLINE_THRESHOLD_MINUTES: int = 1440


class DeviceRepository(AsyncRepository[Device]):
    def __init__(self):
        super().__init__(Device)

    async def get_by_id(self, session: AsyncSession, id: str) -> Optional[Device]:
        # pylint: disable=unexpected-keyword-arg
        return await self.get(session, Device.id == id)

    async def list_by_capability(self, session: AsyncSession, capability: str) -> Sequence[Device]:
        # pylint: disable=no-member,unexpected-keyword-arg
        return await self.list(session, Device.capabilities.contains(capability))

    async def list_offline(self, session: AsyncSession, offline_threshold_minutes: int = OFFLINE_THRESHOLD_MINUTES) -> Sequence[Device]:
        # pylint: disable=no-member
        return await self.list(session, Device.last_seen < datetime.now(timezone.utc) - timedelta(minutes=offline_threshold_minutes), order_by=Device.last_seen.asc())



from typing import Generic, List, Sequence, Type, TypeVar

from fastapi import APIRouter, Body, Depends, HTTPException, Query, status, Response, Path
from sqlalchemy.ext.asyncio import AsyncSession
from sqlmodel import SQLModel

from ..data.repositories import AsyncRepository
from ..data import async_session

T = TypeVar('T', bound=SQLModel)


class BaseRouter(Generic[T]):
    def __init__(self, model: Type[T], repo: AsyncRepository[T], prefix: str, tags: List[str]):
        self.model  = model
        self.repo   = repo
        self.router = APIRouter(prefix=prefix, tags=tags)

        self._setup_routes()

    @staticmethod
    async def _get_session():
        async with async_session() as session:
            yield session

    async def _before_add(self, entity: T, session: AsyncSession) -> None:
        pass

    # pylint: disable=unused-argument
    async def _after_add(self, entity: T, session: AsyncSession) -> None:
        await session.commit()

    async def _before_delete(self, entity: T, session: AsyncSession) -> None:
        pass

    # pylint: disable=unused-argument
    async def _after_delete(self, entity: T, session: AsyncSession) -> None:
        await session.commit()

    # pylint: disable=unused-argument
    async def _before_update(self, entity: T, data: dict, session: AsyncSession) -> None:
        pass

    # pylint: disable=unused-argument
    async def _after_update(self, entity: T, session: AsyncSession) -> None:
        await session.commit()

    def _setup_routes(self) -> None:
        @self.router.post('/', response_model=self.model, status_code=status.HTTP_201_CREATED)
        async def add(data: dict = Body(...), session: AsyncSession = Depends(BaseRouter._get_session)) -> T:
            entity = self.model(**data)

            await self._before_add(entity, session)
            result = await self.repo.add(session, entity)
            await self._after_add(result, session)

            return result

        @self.router.delete('/{id}', status_code=status.HTTP_204_NO_CONTENT)
        async def delete(id: str = Path(..., min_length=1), session: AsyncSession = Depends(BaseRouter._get_session)) -> None:
            entity = await self.repo.get(session, self.model.id == id)
            if not entity:
                raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

            await self._before_delete(entity, session)
            await self.repo.delete(session, entity)
            await self._after_delete(entity, session)

        @self.router.get('/{id}', response_model=self.model)
        async def get(id: str = Path(..., min_length=1), session: AsyncSession = Depends(BaseRouter._get_session)) -> T:
            entity = await self.repo.get(session, self.model.id == id)
            if not entity:
                raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

            return entity

        @self.router.get('/', response_model=List[self.model])
        async def list(response: Response, session: AsyncSession = Depends(BaseRouter._get_session), order_by: str = Query(None), offset: int = Query(0, ge=0), limit: int = Query(10000, ge=1, le=10000)) -> Sequence[T]:
            response.headers['X-Total-Count'] = str(await self.repo.count(session))

            return await self.repo.list(session, order_by=order_by, offset=offset, limit=limit)

        @self.router.put('/{id}', response_model=self.model)
        async def update(id: str = Path(..., min_length=1), data: dict = Body(...), session: AsyncSession = Depends(BaseRouter._get_session)) -> T:
            entity = await self.repo.get(session, self.model.id == id)
            if not entity:
                raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

            await self._before_update(entity, data, session)

            for key, value in data.items():
                setattr(entity, key, value)

            result = await self.repo.update(session, entity)
            await self._after_update(result, session)

            return result



from typing import Optional

from sqlmodel import and_

from ..models import OutboxEvent
from .base_repository import AsyncRepository


class OutboxRepository(AsyncRepository[OutboxEvent]):
    def __init__(self):
        super().__init__(OutboxEvent)

    async def get_next(self, session, device_id: str, event_type: str) -> Optional[OutboxEvent]:
        # pylint: disable=no-member,singleton-comparison
        events = await self.list(session, and_(OutboxEvent.device_id == device_id, OutboxEvent.event_type == event_type, OutboxEvent.is_processed == False), order_by=OutboxEvent.created_at.desc())
        return events[0] if events else None

    async def delete_pending(self, session, device_id: str, event_type: str) -> None:
        # pylint: disable=singleton-comparison
        events = await self.list(session, and_(OutboxEvent.device_id == device_id, OutboxEvent.event_type == event_type, OutboxEvent.is_processed == False))
        for event in events:
            await self.delete(session, event)
